"""Main entry point for the service."""
import os
from os import path
import logging.config
import yaml

from rest_model_service.configuration import Configuration
from rest_model_service.helpers import create_app


logger = logging.getLogger(__name__)


# getting the path to the configuration file
# defaults to "rest_config.yaml" in the current directory if the path not provided
if "REST_CONFIG" in os.environ:
    file_path = os.environ["REST_CONFIG"]
else:
    file_path = "rest_config.yaml"

# loading the configuration
if path.exists(file_path) and path.isfile(file_path):
    with open(file_path) as file:
        configuration_dict = yaml.full_load(file)
    configuration = Configuration(**configuration_dict)
    app = create_app(configuration, wait_for_model_creation=False)
else:
    # if there is no configuration file, or it is not found, then raise a warning
    raise ValueError("Could not find configuration file '{}', service has no models loaded.".format(file_path))



"""Script that generates an OpenAPI schema."""
import os
import sys
import traceback
import argparse
import yaml

from rest_model_service.helpers import create_app
from rest_model_service.configuration import Configuration


def main() -> None:
    """Entry point for the cli tool."""
    argument_parser = argparse.ArgumentParser(description="Generate an OpenAPI schema.")
    argument_parser.add_argument("--configuration_file",  type=str, help="Path of configuration file.",
                                 default="rest_config.yaml")
    argument_parser.add_argument("--output_file", type=str, help="Path of output file.", default="openapi.yaml")

    args = argument_parser.parse_args()

    try:
        with open(args.configuration_file, "r") as file:
            configuration_dict = yaml.full_load(file)
        configuration = Configuration(**configuration_dict)

        # create application object, waiting for model creation to finish in order to have all the model endpoints in
        # the OpenAPI document
        app = create_app(configuration, wait_for_model_creation=True)

        # save OpenAPI document
        with open(args.output_file, "w") as file:
            yaml.dump(app.openapi(), file)
    except Exception:
        traceback.print_exc()
        sys.exit(os.EX_SOFTWARE)
    else:
        sys.exit(os.EX_OK)


if __name__ == "__main__":
    main()



"""Health status manager that holds health status information for the application."""
from typing import Tuple, Dict
from rest_model_service.schemas import HealthStatus, ReadinessStatus, StartupStatus


class StatusManager(object):
    """Health status manager singleton."""

    def __new__(cls, *args: Tuple, **kwargs: Dict):  # noqa: D102, ANN101, ANN204
        """Create new StatusManager instance, after instance is first created it will always be returned."""
        if not hasattr(cls, "_instance"):
            cls._instance = super(StatusManager, cls).__new__(cls, *args, **kwargs)
            cls._instance._is_initialized = False
        return cls._instance

    def __init__(self) -> None:  # noqa: ANN101
        """Construct StatusManager object."""
        if not self._is_initialized:  # pytype: disable=attribute-error
            self._health_status = HealthStatus.UNKNOWN
            self._readiness_status = ReadinessStatus.UNKNOWN
            self._startup_status = StartupStatus.UNKNOWN
            self._is_initialized = True

    @classmethod
    def clear_instance(cls) -> None:  # noqa: ANN102
        """Clear singleton instance from class."""
        if hasattr(cls, "_instance"):
            del cls._instance

    def set_health_status(self, health_status: HealthStatus) -> None:  # noqa: ANN101
        """Set health status of the service."""
        self._health_status = health_status

    def get_health_status(self) -> HealthStatus:  # noqa: ANN101
        """Get health status of the service."""
        return self._health_status

    def set_readiness_status(self, readiness_status: ReadinessStatus) -> None:  # noqa: ANN101
        """Set readiness status of the service."""
        self._readiness_status = readiness_status

    def get_readiness_status(self) -> ReadinessStatus:  # noqa: ANN101
        """Get readiness status of the service."""
        return self._readiness_status

    def set_startup_status(self, startup_status: StartupStatus) -> None:  # noqa: ANN101
        """Set startup status of the service."""
        self._startup_status = startup_status

    def get_startup_status(self) -> StartupStatus:  # noqa: ANN101
        """Get startup status of the service."""
        return self._startup_status



"""Routes for the service."""
import logging
from fastapi import APIRouter
from fastapi.responses import JSONResponse
from starlette.responses import RedirectResponse

from ml_base import MLModel
from ml_base.ml_model import MLModelSchemaValidationException
from ml_base.utilities import ModelManager

from rest_model_service.schemas import ModelDetailsCollection, ModelMetadata, Error
from rest_model_service.status_manager import StatusManager
from rest_model_service.schemas import HealthStatus, ReadinessStatus, StartupStatus, HealthStatusResponse, \
    ReadinessStatusResponse, StartupStatusResponse

logger = logging.getLogger(__name__)


router = APIRouter()


@router.get("/")
async def get_root():   # noqa: ANN201
    """Root of API, redirects to the /docs endpoint."""
    response = RedirectResponse(url='/docs')
    return response


@router.get("/api/health",
            response_model=HealthStatusResponse,
            responses={
                200: {"model": HealthStatusResponse},
                503: {"model": HealthStatusResponse}
            })
async def health_check() -> JSONResponse:   # noqa: ANN201
    """Check on service health.

    Indicates whether the service process is running. This endpoint will return a 200 status once the service
    has started.

    """
    health_status_manager = StatusManager()
    health_status = health_status_manager.get_health_status()
    health_status_response = HealthStatusResponse(health_status=health_status)

    if health_status == HealthStatus.HEALTHY:
        return JSONResponse(health_status_response.dict(), 200)
    else:
        return JSONResponse(health_status_response.dict(), 503)


@router.get("/api/health/ready",
            response_model=ReadinessStatusResponse,
            responses={
                200: {"model": ReadinessStatusResponse},
                503: {"model": ReadinessStatusResponse}
            })
async def readiness_check() -> JSONResponse:   # noqa: ANN201
    """Check on service readiness.

    Indicates whether the service is ready to respond to requests. This endpoint will return a 200 status only if
    all the models and decorators have finished being instantiated without errors. Once the models and decorators
    are loaded, the readiness check will always return a ACCEPTING_TRAFFIC state.

    """
    health_status_manager = StatusManager()
    readiness_status = health_status_manager.get_readiness_status()
    readiness_status_response = ReadinessStatusResponse(readiness_status=readiness_status)

    if readiness_status == ReadinessStatus.ACCEPTING_TRAFFIC:
        return JSONResponse(readiness_status_response.dict(), 200)
    else:
        return JSONResponse(readiness_status_response.dict(), 503)


@router.get("/api/health/startup",
            response_model=StartupStatusResponse,
            responses={
                200: {"model": StartupStatusResponse},
                503: {"model": StartupStatusResponse}
            })
async def startup_check() -> JSONResponse:   # noqa: ANN201
    """Check on service startup.

    Indicates whether the service is started. This endpoint will return a 200 status only if all the models
    and decorators have finished being instantiated without errors.

    """
    health_status_manager = StatusManager()
    startup_status = health_status_manager.get_startup_status()
    startup_status_response = StartupStatusResponse(startup_status=startup_status)

    if startup_status == StartupStatus.STARTED:
        return JSONResponse(startup_status_response.dict(), 200)
    else:
        return JSONResponse(startup_status_response.dict(), 503)


@router.get("/api/models",
            response_model=ModelDetailsCollection,
            responses={
                200: {"model": ModelDetailsCollection},
                500: {"model": Error}
            })
async def get_models() -> JSONResponse:   # noqa: ANN201
    """List of models available.

    This endpoint returns details about all the models currently loaded in the service, however not all models
    necessarily have an endpoint created for them.

    """
    try:
        model_manager = ModelManager()
        model_details_collection = model_manager.get_models()
        model_details_collection = ModelDetailsCollection(**{"models": model_details_collection}).dict()
        return JSONResponse(status_code=200, content=model_details_collection)
    except Exception as e:
        error = Error(type="ServiceError", messages=[str(e)]).dict()
        return JSONResponse(status_code=500, content=error)


@router.get("/api/models/{model_qualified_name}/metadata",
            response_model=ModelMetadata,
            responses={
                200: {"model": ModelMetadata},
                500: {"model": Error}
            })
async def get_model_metadata(model_qualified_name: str) -> JSONResponse:   # noqa: ANN201
    """Return metadata about a single model.

    This endpoint returns metadata about any of the models currently loaded in the service, however not all models
    necessarily have an endpoint created for them.

    """
    try:
        model_manager = ModelManager()
        model_metadata = model_manager.get_model_metadata(qualified_name=model_qualified_name)
        model_metadata = ModelMetadata(**model_metadata).dict()
        return JSONResponse(status_code=200, content=model_metadata)
    except Exception as e:
        error = Error(type="ServiceError", messages=[str(e)]).dict()
        return JSONResponse(status_code=500, content=error)


class PredictionController(object):
    """Callable class that hosts an instance of a model.

    .. note::
       This class is designed to be used as a route in the FastAPI application.

    """

    def __init__(self, model: MLModel) -> None:  # noqa: ANN101
        """Initialize the controller."""
        self._model = model

    def __call__(self, data) -> JSONResponse:  # noqa: ANN001,ANN204,ANN101
        """Make a prediction with a model."""
        try:
            prediction = self._model.predict(data).dict()
            logger.debug("Made a prediction with model '{}'.".format(self._model.qualified_name))
            return JSONResponse(status_code=200, content=prediction)
        except MLModelSchemaValidationException as e:
            logger.exception("Error when making a prediction  prediction with model '{}'.".
                             format(self._model.qualified_name), exc_info=e)
            error = Error(type="SchemaValidationError", messages=[str(e)]).dict()
            return JSONResponse(status_code=400, content=error)
        except Exception as e:
            logger.exception("Error when making a prediction  prediction with model '{}'.".
                             format(self._model.qualified_name), exc_info=e)
            error = Error(type="ServiceError", messages=[str(e)]).dict()
            return JSONResponse(status_code=500, content=error)



"""Schemas used by the service."""
from typing import List, Dict
from enum import Enum
from pydantic import BaseModel, Field


class HealthStatus(str, Enum):
    """Health status of the service."""

    HEALTHY = "HEALTHY"
    NOT_HEALTHY = "NOT_HEALTHY"
    UNKNOWN = "UNKNOWN"


class ReadinessStatus(str, Enum):
    """Readiness status of the service."""

    ACCEPTING_TRAFFIC = "ACCEPTING_TRAFFIC"
    REFUSING_TRAFFIC = "REFUSING_TRAFFIC"
    UNKNOWN = "UNKNOWN"


class StartupStatus(str, Enum):
    """Startup status of the service."""

    STARTED = "STARTED"
    NOT_STARTED = "NOT_STARTED"
    UNKNOWN = "UNKNOWN"


class HealthStatusResponse(BaseModel):
    """Health status response."""

    health_status: HealthStatus = Field(description="Health status of the service.")


class ReadinessStatusResponse(BaseModel):
    """Readiness status response."""

    readiness_status: ReadinessStatus = Field(description="Readiness status of the service.")


class StartupStatusResponse(BaseModel):
    """Startup status response."""

    startup_status: StartupStatus = Field(description="Startup status of the service.")


class ModelDetails(BaseModel):
    """Metadata of a model."""

    display_name: str = Field(description="The display name of the model.")
    qualified_name: str = Field(description="The qualified name of the model.")
    description: str = Field(description="The description of the model.")
    version: str = Field(description="The version of the model.")


class ModelDetailsCollection(BaseModel):
    """Collection of model details."""

    models: List[ModelDetails] = Field(description="Collection of model details.")


class ModelMetadata(ModelDetails):
    """Metadata of a model, includes all information in ModelDetails plus input and output schemas of the model."""

    input_schema: Dict = Field(description="Input schema of a model, as a JSON Schema object.")
    output_schema: Dict = Field(description="Output schema of a model, as a JSON Schema object.")


class Error(BaseModel):
    """Error details."""

    type: str = Field(description="The type of error.")
    messages: List[str] = Field(description="List of error messages.")



"""Exception handler."""
import logging
from starlette.responses import JSONResponse
from fastapi import Request
from fastapi.exceptions import RequestValidationError
from pydantic import ValidationError
from pydantic.error_wrappers import ErrorWrapper

from rest_model_service.schemas import Error


async def validation_exception_handler(request: Request, exception: RequestValidationError) -> JSONResponse:
    """Exception handler."""
    logger = logging.getLogger(__name__)

    messages = []
    for raw_error in exception.raw_errors:
        raw_error: ErrorWrapper
        if type(raw_error.exc) is ValidationError:
            for field_error in raw_error.exc.errors():
                message = "Field '{}' in type '{}' has error '{}', {}.".format(
                    ", ".join(field_error["loc"]),
                    raw_error.exc.model.__name__,
                    field_error["type"],
                    field_error["msg"]
                )
                messages.append(message)
        else:
            messages.append(str(raw_error.exc))

    extra = {
        "action": "predict",
        "endpoint": str(request.url),
        "status": "error",
        "error_info": ", ".join(messages)
    }
    logger.error("Request validation error.", extra=extra)

    error = Error(type="ValidationError", messages=messages).dict()
    return JSONResponse(status_code=400, content=error)



"""RESTful service for hosting machine learning models."""

from os.path import abspath, dirname, join

__all__ = ["__version__"]


try:
    with open(join(abspath(dirname(__file__)), "version.txt"), encoding="utf-8") as f:
        __version__ = f.read()
except Exception:
    __version__ = "N/A"



"""Helper functions."""
import logging
from typing import Type
import importlib
from concurrent.futures import ThreadPoolExecutor, Future
import logging.config
from fastapi import FastAPI
from fastapi.exceptions import RequestValidationError
from ml_base.utilities import ModelManager

from rest_model_service.status_manager import StatusManager, HealthStatus, StartupStatus, ReadinessStatus
from rest_model_service.configuration import Configuration
from rest_model_service.routes import PredictionController  # noqa: F401,E402
from rest_model_service.exception_handlers import validation_exception_handler
from rest_model_service.schemas import Error
from rest_model_service.routes import router

logger = logging.getLogger(__name__)


def load_type(class_path: str) -> Type:
    """Load a type.

    Args:
        class_path: Path of the type.

    Returns:
        Type

    """
    # splitting the class_path into module path and class name
    module_path = ".".join(class_path.split(".")[:-1])
    class_name = class_path.split(".")[-1]

    # importing the model class
    _module = importlib.import_module(module_path)
    _class = getattr(_module, class_name)
    return _class


def create_app(configuration: Configuration, wait_for_model_creation: bool = False) -> FastAPI:
    """Create instance of FastAPI app and return it.

    Args:
        configuration: Configuration used to create application
        wait_for_model_creation: Whether to wait for models to finish instantiating before returning. Defaults to false
            to allow for asynchronous model creation in a separate thread.

    Returns:
        FastAPI application object.

    .. note::
        This function creates a FastAPI application object using the options in a Configuration object. The application
        scaffolding is created by this function. The models are added to the application asynchronously by calling the
        build_models function in a separate thread.

    """
    logger.info("Creating FastAPI app for: '{}'.".format(configuration.service_title))

    app: FastAPI = FastAPI(title=configuration.service_title,
                           description=configuration.description,
                           version=configuration.version)

    # adding a custom exception handler for validation errors
    app.add_exception_handler(RequestValidationError, validation_exception_handler)

    # adding common endpoints to the app
    app.include_router(router)

    # setting the health, startup, and readiness status of the application, the app remains in the "HEALTHY",
    # "REFUSING_TRAFFIC", and "NOT_STARTED" state until all models and decorators are initialized.
    health_status_manager = StatusManager()
    health_status_manager.set_health_status(HealthStatus.HEALTHY)
    health_status_manager.set_readiness_status(ReadinessStatus.REFUSING_TRAFFIC)
    health_status_manager.set_startup_status(StartupStatus.NOT_STARTED)

    # creating models and decorators in a separate thread
    executor = ThreadPoolExecutor(max_workers=1)
    future = executor.submit(build_models, app, configuration)

    # add the set_service_status callback
    future.add_done_callback(set_service_status)
    executor.shutdown(wait=False)  # shutting down the executor asynchronously after the task finishes

    # if waiting for model creation, then wait for result here instead of returning app object immediately
    # this is useful for creating the app object for tests and for getting the full OpenAPI spec from the app
    if wait_for_model_creation:
        _ = future.result()
    return app


def set_service_status(future: Future) -> None:
    """Set the status of the service according to the results of the build_models thread.

    Args:
        future: Future for the build_models task.

    Returns:
        None

    .. note::
        If the models or decorators created in the build_models function raise an exception, the status of the service
        will be "NOT_HEALTHY", "REFUSING_TRAFFIC", and "STARTED".

    """
    health_status_manager = StatusManager()

    if future.exception() is not None:
        health_status_manager.set_health_status(HealthStatus.NOT_HEALTHY)
        health_status_manager.set_readiness_status(ReadinessStatus.REFUSING_TRAFFIC)
        health_status_manager.set_startup_status(StartupStatus.STARTED)
    else:
        health_status_manager.set_health_status(HealthStatus.HEALTHY)
        health_status_manager.set_readiness_status(ReadinessStatus.ACCEPTING_TRAFFIC)
        health_status_manager.set_startup_status(StartupStatus.STARTED)


def build_models(app: FastAPI, configuration: Configuration) -> None:
    """Instantiate models and decorators, adding endpoints if necessary.

    Args:
        app: FastAPI app to modify by adding model endpoints.
        configuration: Configuration to use to instantiate models and decorators.

    Returns:
        Optional exception object, if an exception is raised it is returned to the caller.

    .. note::
        This function is designed to execute as a separate thread from the main server thread. This allows the FastAPI
        server to start up and return status information through the health endpoints before all the models and
        decorators have finished being instantiated.

    """
    # setting up the logging configuration
    if configuration.logging is not None:
        logging.config.dictConfig(configuration.logging)

    # loading the models into the ModelManager singleton instance
    model_manager = ModelManager()

    for model in configuration.models:
        # loading the model's class
        model_class = load_type(model.class_path)

        # instantiating the model object from the class
        model_instance = model_class()

        # adding the model instance to the ModelManager
        model_manager.add_model(model_instance)
        logger.info("Loaded {} model.".format(model_instance.qualified_name))

        decorators = model.decorators if model.decorators is not None else []
        # initializing decorators for the model
        for decorator in decorators:
            # loading the decorator's class
            decorator_class = load_type(decorator.class_path)

            # instantiating the decorator object from the class
            if decorator.configuration is not None:
                decorator_instance = decorator_class(**decorator.configuration)
            else:
                decorator_instance = decorator_class()

            # adding the decorator to the model in the ModelManager
            model_manager.add_decorator(model.qualified_name, decorator_instance)

            logger.info("Added {} decorator to {} model.".format(decorator_class.__name__,
                                                                 model_instance.qualified_name))

        # creating an endpoint for each model, if the configuration allows it
        if model.create_endpoint:
            model = model_manager.get_model(model.qualified_name)

            controller = PredictionController(model=model)
            controller.__call__.__annotations__["data"] = model.input_schema

            app.add_api_route("/api/models/{}/prediction".format(model.qualified_name),
                              controller,
                              methods=["POST"],
                              response_model=model.output_schema,
                              description=model.description,
                              responses={
                                  400: {"model": Error},
                                  500: {"model": Error}
                              })
            logger.info("Created endpoint for {} model.".format(model.qualified_name))
        else:
            logger.info("Skipped creating an endpoint for model: {}".format(model.qualified_name))
